import os
import scipy
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.preprocessing import Normalizer
from sklearn.decomposition import PCA

SITE = 'site'
DISH = 'dish'
TOTAL = 'total'

class SingleCellPCAAnalysis:
    def __init__(self, root: str, conditions: list):
        self.root = root
        self.conditions = conditions
        self.decay_matrices = {}         
        self.decay_features = pd.DataFrame()
        self.nadh_decay = pd.DataFrame()
        self.fad_decay = pd.DataFrame()
        self.decay_multi_channel = pd.DataFrame()
        self.pca_model = None            
        self.explained_variance = None  

    def load_decay_data(self):
        for condition in self.conditions:
            condition_path = os.path.join(self.root, condition)
            if not os.path.exists(condition_path):
                continue
            for f in os.listdir(condition_path):
                path = os.path.join(condition_path, f)
                if not os.path.isdir(path):
                    continue
                decay_files = [name for name in os.listdir(path) if '_decay_matrix.npy' in name]
                for df in decay_files:
                    full_decay_path = os.path.join(path, df)
                    matrix = np.load(full_decay_path)
                    self.decay_matrices[full_decay_path] = matrix

    def build_features_dataframe(self):
        decay_features_list = []
        for path, matrix_data in self.decay_matrices.items():
            info = path.split(os.sep)[-1].split("_")
            df = pd.DataFrame(matrix_data)
            df.insert(loc=0, column='fullName', value=path)
            df.insert(loc=1, column='treatment', value=info[1])
            df.insert(loc=2, column='dish', value=info[2])
            df.insert(loc=3, column='site', value=info[3][0])
            df.insert(loc=4, column='Channel', value=info[3][-1])
            df.insert(loc=5, column='cell_id', value=np.arange(matrix_data.shape[0]))
            decay_features_list.append(df)

        self.decay_features = pd.concat(decay_features_list).reset_index(drop=True)

    def separate_nadh_fad(self):
        self.nadh_decay = self.decay_features[self.decay_features.Channel == 'n'].copy()
        self.fad_decay = self.decay_features[self.decay_features.Channel == 'f'].copy()

    def normalize_data(self):
        # NADH
        nadh_cols = [col for col in self.nadh_decay.columns if isinstance(col, int)]
        if nadh_cols:
            F_nadh = self.nadh_decay.loc[:, nadh_cols]
            normalized_n = Normalizer(norm='l1').fit_transform(F_nadh)
            self.nadh_decay.loc[:, nadh_cols] = normalized_n

        # FAD
        fad_cols = [col for col in self.fad_decay.columns if isinstance(col, int)]
        if fad_cols:
            F_fad = self.fad_decay.loc[:, fad_cols]
            normalized_f = Normalizer(norm='l1').fit_transform(F_fad)
            self.fad_decay.loc[:, fad_cols] = normalized_f

    def merge_nadh_fad(self):
        merged = pd.merge(
            self.nadh_decay,
            self.fad_decay,
            on=["treatment", "dish", "site", "cell_id"],
            how="inner"
        )

        to_drop = ["fullName_x", "fullName_y", "index_x", "Channel_x", "index_y", "Channel_y"]
        for c in to_drop:
            if c in merged.columns:
                merged.drop(columns=c, inplace=True)

        col_names = ["treatment", "dish", "site", "cell_id"] + [x for x in range(256 * 2)]
        if len(merged.columns) == len(col_names):
            merged.columns = col_names

        self.decay_multi_channel = merged

    def pca_plot_any(self, df, comp_x=0, comp_y=1, title="Scatter Plot"):
        if df.empty:
            print("DataFrame is empty, cannot perform PCA.")
            return None, None, None, None

        if 'treatment' not in df.columns:
            print("DataFrame has no 'treatment' column, cannot color by treatment.")
            return None, None, None, None

        numeric_cols = [col for col in df.columns if isinstance(col, int)]
        if not numeric_cols:
            print("No numeric columns to perform PCA on.")
            return None, None, None, None

        features_data = df[numeric_cols]

        pca_model = PCA()
        X = pca_model.fit_transform(features_data)
        explained_variance = pca_model.explained_variance_ratio_

        plot_df = pd.concat([
            df[['treatment']].reset_index(drop=True),
            pd.DataFrame(X)
        ], axis=1)
        plot_df = plot_df.rename(columns={
            comp_x: f'PC{comp_x}',
            comp_y: f'PC{comp_y}'
        })

        sns.scatterplot(
            data=plot_df,
            x=f'PC{comp_x}',
            y=f'PC{comp_y}',
            hue='treatment',
            palette=sns.color_palette('Set2'),
            s=12,
            alpha=0.75
        )
        plt.title(title)
        plt.legend(title='Treatment')
        plt.show()

        print(f"Explained Variance Ratio (PC{comp_x+1}, PC{comp_y+1}):")
        for i in range(min(len(explained_variance), 10)):
            print(f"PC{i+1}: {explained_variance[i]:.4f}")

        return plot_df, pca_model, features_data, explained_variance

    def pca_plot(self, comp_x=0, comp_y=1, title="Scatter Plot"):
        if self.decay_multi_channel.empty:
            print("No data in self.decay_multi_channel. Please run merge_nadh_fad() first.")
            return None, None, None, None

        plot_df, pca_model, features_data, explained_variance = self.pca_plot_any(
            self.decay_multi_channel, comp_x, comp_y, title
        )
        self.pca_model = pca_model
        self.explained_variance = explained_variance

        return plot_df, pca_model, features_data, explained_variance

    def anova_analysis(self, data: pd.DataFrame) -> pd.DataFrame:
        if 'treatment' not in data.columns:
            raise ValueError("Data does not contain 'treatment' column.")

        treatments = data['treatment'].unique()
        anova_results = []
        drop_cols = ["treatment"]
        use_cols = [c for c in data.columns if c not in drop_cols]

        for col in use_cols:
            groups = [data[data['treatment'] == treatment][col] for treatment in treatments]
            f_statistic, p_value = scipy.stats.f_oneway(*groups)
            anova_results.append({"statistic": f_statistic, "pvalue": p_value})

        anova_df = pd.DataFrame(anova_results, index=use_cols)
        anova_df.sort_values("statistic", ascending=False, inplace=True)
        return anova_df

    def run_pipeline(self):
        self.load_decay_data()
        self.build_features_dataframe()
        self.separate_nadh_fad()
        self.normalize_data()
        self.merge_nadh_fad()
        self.pca_plot()




import skimage.io
import matplotlib.pyplot as plt
import numpy as np
import os
import re
from sdtfile.sdtfile import SdtFile
import torch
import pandas as pd
from sdt_reading import read_sdt150 

gpu = 3
device = f"cuda:{gpu}" if torch.cuda.is_available() else 'cpu'
    

class SingleCellFeatureExtractor:
    def __init__(self, path, image_type='n'):
        self.path = path
        self.image_type = image_type.lower()
        if self.image_type == 'n':
            self.channel_2ch = 1
            self.channel_1ch = 0
            self.default_ymax = 25000
        else:
            self.channel_2ch = 0
            self.channel_1ch = 0
            self.default_ymax = 6000
        self.features_dict = {}

    def load_features(self):
        d = {}

        df = pd.read_csv(self.path)

        list_file_path = df["file_path"].tolist()
        list_channel_type = df["channel_type"].tolist()

        for i in range(len(list_file_path)):
            if (list_channel_type[i] != self.image_type):
                continue
            fp = os.path.join(list_file_path[i])
            sdt_files = [nm for nm in os.listdir(fp) if nm.endswith('.sdt')]
            if not sdt_files:
                continue

            sdt_full = os.path.join(fp, sdt_files[0])

            arr = read_sdt150(sdt_full)
            if arr.ndim == 3:
                arr = arr[None, ...]  # ensure shape is (1, X, Y, T)
            d[sdt_full] = arr

        self.features_dict = d

    @staticmethod
    def compute_matrix(raw, masks, channel):
        f = []
        u = np.unique(masks)
        for c in u[u > 0]:
            m = masks == c
            if not np.any(m):
                continue
            mc = raw[channel] * m[:, :, None]
            r, col = np.where(m)
            mnr, mxr = np.min(r), np.max(r)
            mnc, mxc = np.min(col), np.max(col)
            sc = mc[mnr:mxr+1, mnc:mxc+1]
            t = np.sum(sc, axis=(0, 1))
            f.append(t)
        return np.asarray(f)

    def plot_features(self, matrix, y_max=None):
        if y_max is None:
            y_max = self.default_ymax
        print("Matrix shape:", matrix.shape)
        for i in range(matrix.shape[0]):
            plt.plot(matrix[i])
        plt.ylim([0, y_max])
        plt.show()

    def process_features(self):
        for sdt_path, raw in self.features_dict.items():
            if raw.shape[0] == 2:
                ch = self.channel_2ch
            else:
                ch = self.channel_1ch

            if self.image_type == 'n':
                mask_path = sdt_path.replace(".sdt", "_photons_cyto3.tiff")
            elif self.image_type == 'f':
                mask_path = re.sub(r'(_\d)f', r'\1n', sdt_path)
                mask_path = mask_path.replace(".sdt", "_photons_cyto3.tiff")
            else:
                mask_path = sdt_path.replace(".sdt", "_photons_cyto3.tiff")

            if not os.path.exists(mask_path):
                print(f"Skipping {sdt_path}, mask not found: {mask_path}")
                continue

            m = skimage.io.imread(mask_path)
            mat = self.compute_matrix(raw, m, ch)
            self.plot_features(mat, None)
            out = sdt_path.replace(".sdt", "_feature_matrix.npy")
            np.save(out, mat)
            print(f"Saved feature matrix: {out}")

    def run(self):
        self.load_features()
        self.process_features()


from setuptools import setup, find_packages

setup(
    name="skala_lab_package",
    version="0.1.1",
    author="Manato Ogawa",
    author_email="maogawa@wisc.edu",
    description="A package for redox feature analysis, CSV generation, and SDT file handling.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/monatopotato",  
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "numpy",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




import os
import numpy as np
import pandas as pd
import ot
import matplotlib.pyplot as plt
from sklearn.preprocessing import Normalizer
from sklearn.decomposition import PCA

class Optimal_Transport:
    def __init__(self, root, conditions):
        self.root = root
        self.conditions = conditions
        self.decay_features = None
        self.nadh_decay = None
        self.fad_decay = None
        self.decay_multi_channel_features = None
        self.pca = None
        self.normalizer = None
        self.results = {}
    
    def load_decay_data(self):
        cytoplasm_decay_matrices = {}
        for condition in self.conditions:
            files = os.listdir(os.path.join(self.root, condition))
            for f in files:
                path = os.path.join(self.root, condition, f)
                if path[-2] == "1" and "Dish1" in path and "Control" in path:
                    continue
                decay_file = [name for name in os.listdir(path) if '_decay_matrix.npy' in name and 'cytoplasm' not in name]
                if not decay_file:
                    continue
                decay_path = os.path.join(path, decay_file[0])
                cytoplasm_decay_matrices[decay_path] = np.load(decay_path)
        
        decay_features = []
        for k, v in cytoplasm_decay_matrices.items():
            df = pd.DataFrame(v)
            info = os.path.basename(k).split("_")
            df.insert(0, 'fullName', k)
            df.insert(1, 'treatment', os.path.basename(os.path.dirname(os.path.dirname(k))))  
            df.insert(2, 'dish', info[2])
            df.insert(3, 'site', info[3][0])
            df.insert(4, 'Channel', info[3][-1])
            df.insert(5, 'cell_id', np.arange(v.shape[0]))
            decay_features.append(df)
        
        self.decay_features = pd.concat(decay_features).reset_index()
        self.nadh_decay = self.decay_features[self.decay_features.Channel == "n"].copy()
        self.fad_decay = self.decay_features[self.decay_features.Channel == "f"].copy()
        
        # self.nadh_decay.iloc[:, 7:] = Normalizer(norm="l1").fit_transform(self.nadh_decay.iloc[:, 7:])
        # self.fad_decay.iloc[:, 7:] = Normalizer(norm="l1").fit_transform(self.fad_decay.iloc[:, 7:])
        
        # self.decay_multi_channel_features = pd.merge(self.nadh_decay, self.fad_decay, on=["treatment", "dish", "site", "cell_id"], how="inner")
        # self.decay_multi_channel_features.drop(["fullName_x", "fullName_y", "index_x","Channel_x","index_y","Channel_y"], axis=1, inplace=True)
        # self.decay_multi_channel_features.columns = ["treatment", "dish", "site", "cell_id"] + [str(x) for x in range(256*2)]
    
    def compute_global_pca(self, n_components=10):
        X_all = self.nadh_decay.iloc[:, 7:].values
        self.normalizer = Normalizer(norm='l1')
        X_all_norm = self.normalizer.fit_transform(X_all)
        self.pca = PCA(n_components=n_components)
        self.pca.fit(X_all_norm)
    
    def get_ot_distance(self, treat_control, treat_experiment):
        df_group1 = self.nadh_decay[self.nadh_decay['treatment'] == treat_control]
        df_group2 = self.nadh_decay[self.nadh_decay['treatment'] == treat_experiment]
        
        if df_group1.empty or df_group2.empty:
            print(f"Warning: No cells found for {treat_control} or {treat_experiment}. Skipping.")
            return None
        
        X_group1 = self.normalizer.transform(df_group1.iloc[:, 7:].values)
        X_group2 = self.normalizer.transform(df_group2.iloc[:, 7:].values)
        X_group1_pca = self.pca.transform(X_group1)
        X_group2_pca = self.pca.transform(X_group2)
        
        M = ot.dist(X_group1_pca, X_group2_pca, metric='euclidean')
        M /= M.max()
        
        M1 = ot.dist(X_group1_pca, X_group1_pca, metric='euclidean')
        M1 /= M1.max()
        
        M2 = ot.dist(X_group2_pca, X_group2_pca, metric='euclidean')
        M2 /= M2.max()
        
        G_matrix = ot.emd(np.ones(X_group1_pca.shape[0]) / X_group1_pca.shape[0],
                          np.ones(X_group2_pca.shape[0]) / X_group2_pca.shape[0], M, numItermax=1e7)
        dist = np.sum(G_matrix * M)
        
        self.results[(treat_control, treat_experiment)] = {
            "dist": dist,
            "M": M,
            "M1": M1,
            "M2": M2
        }
        
        return self.results[(treat_control, treat_experiment)]
    
    def plot_ot_distances(self, pairs):
        distances = []
        labels = []
        
        for (control_name, treat_name) in pairs:
            result = self.get_ot_distance(control_name, treat_name)
            if result is not None:
                distances.append(result["dist"])
                labels.append(f"{control_name}\nvs\n{treat_name}")
                print(f"OT distance: {control_name} vs {treat_name} = {result['dist']:.4f}")
        
        plt.figure(figsize=(10, 6))
        plt.bar(labels, distances, color='skyblue')
        plt.ylabel("OT Distance")
        plt.title("Pairwise OT Distances (Control vs Treatment) with L1 Normalization + PCA")
        plt.xticks(rotation=30, ha='right')
        plt.tight_layout()
        plt.show()
    
    def compare_distance_histograms(self, M1, M):
        vals_M1 = M1[np.triu_indices(M1.shape[0], k=1)]
        vals_M = M.flatten()
        
        plt.figure(figsize=(8, 6))
        plt.hist(vals_M1, bins=50, alpha=0.5, label="M1 (within-group)", density=True)
        plt.hist(vals_M, bins=50, alpha=0.5, label="M (between-group)", density=True)
        plt.xlabel("Pairwise Distance")
        plt.ylabel("Density")
        plt.title("Comparison of Distance Distributions")
        plt.legend()
        plt.show()
    
    def compare_distance_histograms_fair(self, M1, M):
        vals_M1 = M1[np.triu_indices(M1.shape[0], k=1)]
        vals_M = M.flatten()
        
        min_samples = min(len(vals_M1), len(vals_M))
        vals_M1_sampled = np.random.choice(vals_M1, size=min_samples, replace=False)
        vals_M_sampled = np.random.choice(vals_M, size=min_samples, replace=False)
        
        plt.figure(figsize=(8, 6))
        plt.hist(vals_M1_sampled, bins=50, alpha=0.5, label="M1 (within-group)", density=True)
        plt.hist(vals_M_sampled, bins=50, alpha=0.5, label="M (between-group)", density=True)
        plt.xlabel("Pairwise Distance")
        plt.ylabel("Density")
        plt.title("Comparison of Distance Distributions (Matched Samples)")
        plt.legend()
        plt.show()
